local LAYER_AIRP = "AIRPORT"
local LAYER_RUNWAY = "RUNWAY"
local LAYER_VOR = "VHF_NAVAID"
local LAYER_NDB = "NDB"
local LAYER_WAYPOINT = "WAYPOINT"
local LAYER_LABEL_STYLE = "font:inconsolata_regular.ttf; size:16; color:cyan; halign:left; valign:center"
local ACTIVE_GPS_LAYER_LABEL_STYLE = "font:inconsolata_regular.ttf; size:16; color:magenta; halign:left; valign:center"
local GPS_LAYER_LABEL_STYLE = "font:inconsolata_regular.ttf; size:16; color:white; halign:left; valign:center"
local BOUNDARY_SIZE = 75
local MIN_RANGE = 2
local MAX_RANGE = 900

local map = nil
local memories = nil
local nav_objects_in_range = {}
local nav_object_timer = nil
local layer_drawing_fns = {}
local map_enabled = false
local selected_memory = 0
local selected_range = 10
local selected_view_centered = true
local viewport_size = 0
local position = {lat = 0, long = 0}
local true_heading = 0

local function aircraft_position_in_canvas()
    return {x = viewport_size / 2, y = fif(selected_view_centered, viewport_size / 2, viewport_size - 25)}
end

local function project_map_object_to_canvas(map_origin, map_location)
    aircraft_pos = aircraft_position_in_canvas()
    pixel_factor = aircraft_pos.y / selected_range
    dist_brg = helpers_distance_and_bearing(map_origin, map_location)
    pixel_dist = dist_brg.distance * pixel_factor
    abs_bearing = true_heading - dist_brg.bearing
    return helpers_get_object_location(aircraft_pos, pixel_dist, abs_bearing)
end

local function draw_airport(x, y, airport)
    radius = 10

    if selected_range <= 15  then
        radius = fif(selected_range < 10, fif(selected_range < 5, fif(selected_range < 3, 50, 40), 30), 20)
        helpers_visit_array(nav_objects_in_range[LAYER_RUNWAY], function (index, runway)
            if runway.AIRPORT_ID == airport.ID then
                log(runway)
                bearing_diff = runway.BEARING_TRUE - true_heading
                reciprocal = fif(bearing_diff < 0, -180, 180)
                abs_diff = math.abs(bearing_diff)
                rotation = fif(abs_diff > 90 and abs_diff < 270, bearing_diff - reciprocal, bearing_diff)
                _rotate(rotation, x, y)
                _move_to(x, y - radius)
                _line_to(x, y + radius)
                if (selected_range < 10) then
                    label_x = x - 8
                    label_y = fif(bearing_diff == rotation, y + radius + 5, y - radius - 5)
                    _txt(string.sub(runway.NAME, 3), LAYER_LABEL_STYLE, label_x, label_y)
                end
                _stroke("cyan", 2)
                _rotate(-rotation, x, y)
            end
        end)
    else
        _circle(x, y, radius)
        _stroke("cyan", 2)
    end
    _txt(airport.ICAO, LAYER_LABEL_STYLE, x + radius + 5, y - radius - 5)
end

local function draw_ndb(x, y, ndb)
        radius = 10
        num_rings = 4
        for i=1, num_rings do
            _circle(x, y, i * radius / num_rings)
        end
        _stroke("cyan", 1)
        _txt(string.sub(ndb.ICAO, 1, 5), LAYER_LABEL_STYLE, x + radius, y - radius)
        _txt(string.format("%s", ndb.FREQUENCY), LAYER_LABEL_STYLE, x + radius, y + radius)
end

local function draw_vor(x, y, vor)
        radius = 10
        p1x = x
        p1y = y - radius
        p2x = helpers_x_position(x, radius, 120)
        p2y = helpers_y_position(y, radius, 120)
        p3x = helpers_x_position(x, radius, 240)
        p3y = helpers_y_position(y, radius, 240)
        _triangle(p1x, p1y, p2x, p2y, p3x, p3y)
        _rect(p1x - 2, p1y - 1, 4, 2)
        _rotate(60, p2x, p2y)
        _rect(p2x - 2, p2y - 1, 4, 2)
        _rotate(-60, p2x, p2y)
        _rotate(-60, p3x, p3y)
        _rect(p3x - 2, p3y - 1, 4, 2)
        _rotate(60, p3x, p3y)
        _stroke("cyan", 2)
        _txt(string.sub(vor.ICAO, 1, 5), LAYER_LABEL_STYLE, x + radius, y - radius)
        _txt(string.format("%s", vor.FREQUENCY), LAYER_LABEL_STYLE, x + radius, y + radius)
end

local function draw_waypoint(x, y, wp)
    radius = 10
    colour = fif(wp.active, "magenta", "white")
    fill = fif(wp.active, "active_gps.png", "gps.png")
    style = fif(wp.active, ACTIVE_GPS_LAYER_LABEL_STYLE, GPS_LAYER_LABEL_STYLE)

    _rect(x - radius,y - radius, 20, 20)
    _fill_img(fill, x - radius,y - radius, 20, 20)
    _txt(wp.id, style, x + radius, y - radius)

    if wp.prev_lat and wp.prev_long then
        prev_pos = project_map_object_to_canvas(position, {lat = wp.prev_lat, long = wp.prev_long})

        _move_to(x, y)
        _line_to(prev_pos.x, prev_pos.y)
        _stroke(colour, 2)
    end
end

local function draw_nav_layer_object(nav_type, nav_obj)
    object_pos = project_map_object_to_canvas(position, {lat = nav_obj.LATITUDE, long = nav_obj.LONGITUDE})
    layer_drawing_fns[nav_type](object_pos.x, object_pos.y, nav_obj)
end

local function draw_nav_layers()
    selected_layers = fif(selected_memory > 0, memories[selected_memory], {})
    max_per_layer = 30

    helpers_visit_array(selected_layers,
        function (index, layer)
            layer_objects = nav_objects_in_range[layer]
            starting_index = fif(#layer_objects > max_per_layer, #layer_objects - max_per_layer, 0)

            helpers_visit_array(layer_objects, function (index, obj)
                if index > starting_index then
                    draw_nav_layer_object(layer, obj)
                end
            end)
        end)
end

local function update_nav_objects_in_range()
    helpers_visit_array({LAYER_AIRP, LAYER_VOR, LAYER_NDB, LAYER_RUNWAY},
        function (index, layer)
            max_range = fif(layer == LAYER_VOR or layer == LAYER_NDB, 400, 200)
            nav_objects = nav_get_radius(layer, position.lat, position.long,
                                         fif(selected_range < max_range, selected_range, max_range),
                100, function (nav_objects)
                    nav_objects_in_range[layer] = nav_objects or {}
                end)
        end)
end

local function redraw()
    visible(map, map_enabled)

    if map_enabled and position then
        canvas_draw(map, draw_nav_layers)
    end

    memory_layers = memories[selected_memory]
    selected_layers = {}

    if memory_layers then
        max_per_layer = 30

        for i = 1, #memory_layers do
            selected_layers[i] = fif(#nav_objects_in_range[memory_layers[i]] < max_per_layer, memory_layers[i],
                "RESTRICTED_" .. memory_layers[i])
        end
    end

    upper_display_update_map_data(map_enabled, selected_memory, selected_layers, selected_range)
end

local function update_nav_object_timer()
    if map_enabled then
        if not nav_object_timer then
            nav_object_timer = timer_start(0, 5000, update_nav_objects_in_range)
        end
    elseif nav_object_timer then
        timer_stop(nav_object_timer)
        nav_object_timer = nil
    end
end

function map_init(centre, canvas_size)
    viewport_size = canvas_size - (2 * BOUNDARY_SIZE)
    map = canvas_add(BOUNDARY_SIZE, BOUNDARY_SIZE, viewport_size, viewport_size)

    layer_drawing_fns[LAYER_AIRP] = draw_airport
    layer_drawing_fns[LAYER_VOR] = draw_vor
    layer_drawing_fns[LAYER_NDB] = draw_ndb
    layer_drawing_fns[LAYER_WAYPOINT] = draw_waypoint

    memories = {
        {LAYER_WAYPOINT},
        {LAYER_AIRP, LAYER_WAYPOINT},
        {LAYER_VOR, LAYER_WAYPOINT},
        {LAYER_NDB, LAYER_WAYPOINT}
    }

    nav_objects_in_range[LAYER_WAYPOINT] = {}
    nav_objects_in_range[LAYER_AIRP] = {}
    nav_objects_in_range[LAYER_RUNWAY] = {}
    nav_objects_in_range[LAYER_VOR] = {}
    nav_objects_in_range[LAYER_NDB] = {}

    redraw()
end

function map_power_on(val)
    map_enabled = val and selected_memory > 0
    redraw()
end

function map_update_location(new_lat, new_long, true_hdg)
    true_heading = true_hdg
    position = {lat = new_lat, long = new_long}
    redraw()
end

function map_update_gps_data(gps_wp_active, gps_direct_to,
                             next_wp_id, next_wp_lat, next_wp_long,
                             prev_wp_valid,
                             prev_wp_id, prev_wp_lat, prev_wp_long)
    wps = {}
    index = 1
    if gps_wp_active then
        if prev_wp_valid then
            wps[index] = {
                id = prev_wp_id,
                LATITUDE = prev_wp_lat,
                LONGITUDE = prev_wp_long,
                active = false,
            }
            index = index + 1
        end

        wps[index] = {
            id = next_wp_id,
            LATITUDE = next_wp_lat,
            LONGITUDE = next_wp_long,
            active = true,
            prev_lat = fif(prev_wp_valid, prev_wp_lat, fif(gps_direct_to, position.lat, nil)),
            prev_long = fif(prev_wp_valid, prev_wp_long, fif(gps_direct_to, position.long, nil))
        }
    end
    nav_objects_in_range[LAYER_WAYPOINT] = wps
    redraw()
end

function map_zoom_in()
    if map_enabled and selected_range > MIN_RANGE then
        if selected_range > 500 then
            selected_range = selected_range - 100
        elseif selected_range > 200 then
            selected_range = selected_range - 50
        elseif selected_range > 100 then
            selected_range = selected_range - 20
        elseif selected_range > 50 then
            selected_range = selected_range - 10
        elseif selected_range > 10 then
            selected_range = selected_range - 5
        elseif selected_range > 4 then
            selected_range = selected_range - 2
        else
            selected_range = selected_range - 1
        end
        update_nav_objects_in_range()
        redraw()
    end
end

function map_zoom_out()
    if map_enabled and selected_range < MAX_RANGE then
        if selected_range < 4 then
            selected_range = selected_range + 1
        elseif selected_range < 10 then
            selected_range = selected_range + 2
        elseif selected_range < 50 then
            selected_range = selected_range + 5
        elseif selected_range < 100 then
            selected_range = selected_range + 10
        elseif selected_range < 200 then
            selected_range = selected_range + 20
        elseif selected_range < 500 then
            selected_range = selected_range + 50
        else
            selected_range = selected_range + 100
        end
        update_nav_objects_in_range()
        redraw()
    end
end

function map_switch_memory()
    selected_memory = fif(selected_memory >= #memories, 0, selected_memory + 1)
    map_enabled = selected_memory > 0

    update_nav_object_timer()
    redraw()
    return map_enabled
end

function map_switch_view()
    selected_view_centered = not selected_view_centered

    redraw()
end

function map_toggle_clear()
    map_enabled = not map_enabled and selected_memory > 0

    update_nav_object_timer()
    redraw()
end
